/*	
 * Test module that creates some files using the proc_ov api
 */ 
 
#include <linux/module.h>

#ifdef CONFIG_SMP
#define __SMP__
#endif

#include <linux/kernel.h>
#include <linux/proc_fs.h>

#define PRCP "test_proc_ov: " 

static int write_simple_string(char *buf, void *context, void *data)
{
	if (context)
		printk(KERN_ERR PRCP
		       "write_simple_string expected NULL context\n");
	strcpy(buf, "fish");
	return 4;
}
static int write_simple_int(int *buf, void *context, void *data)
{
	if (context)
		printk(KERN_ERR PRCP
		       "write_simple_int expected NULL context\n");
	*buf = 5;
	return 0;
}
static int write_simple_enum(int *buf, void *context, void *data)
{
	if (context)
		printk(KERN_ERR PRCP
		       "write_simple_enum expected NULL context\n");
	*buf = 1;
	return 0;
}

static void *simple_context_enter(void *old_context, void *data) 
{
	if (old_context)
		printk(KERN_ERR PRCP 
		       "simple_context_enter expected NULL context\n");
	return (void*) 42;
}
static void simple_context_leave(void *old_context,
				 void *new_context, 
				 void *data) 
{
	if (old_context)
		printk(KERN_ERR PRCP 
		       "simple_context_leave expected NULL old_context\n");
	if (((int)new_context) != 42)
		printk(KERN_ERR PRCP 
		       "simple_context_leave expected 42 new_context\n");
}

static void *subenter42(void *old_context, 
			void *data)
{
	if (((int)old_context) != 42)
		printk(KERN_ERR PRCP 
		       "subenter42 expected 42 context\n");
	return (void*) 43;
}
static void subleave42(void *old_context,
		       void *new_context, 
		       void *data) 
{
	if (((int)old_context) != 42)
		printk(KERN_ERR PRCP 
		       "subleave42 expected 42 old_context\n");
	if (((int)new_context) != 43)
		printk(KERN_ERR PRCP 
		       "subleave42 expected 43 new_context\n");
}

static void *subenter43(void *old_context, 
			void *data)
{
	if (((int)old_context) != 43)
		printk(KERN_ERR PRCP 
		       "subenter43 expected 43 context\n");
	return (void*) 44;
}
static void subleave43(void *old_context,
		       void *new_context, 
		       void *data) 
{
	if (((int)old_context) != 43)
		printk(KERN_ERR PRCP 
		       "subleave43 expected 43 old_context\n");
	if (((int)new_context) != 44)
		printk(KERN_ERR PRCP 
		       "subleave43 expected 44 new_context\n");
}

static int get_context_int(int *buf, void *context, void *data)
{
	*buf = (int) context;
	return 0;
}

static int write_17(int *buf, void *context, void *data)
{
	*buf = 17;
	return 0;
}
static int write_hello(char *buf, void *context, void *data)
{
	strcpy(buf, "hello");
	return strlen("hello");
}
static int write_enum_two(int *buf, void *context, void *data)
{
	*buf = 2;
	return 0;
}


#define DYNDIRNUM 10
static int dyndirc = 0; /* for testing only, known race condition */
int dyn1_num(void *context, void *data) 
{
printk(KERN_ERR PRCP "dyn1_num(%d)\n", (int)context);
	if (((int)context) != 42)
		printk(KERN_ERR PRCP "dyn1_num expected 42 context\n");

	return DYNDIRNUM;
}
int dyn1_name(int index,
	      void *context, 
	      void *data,
	      char *name_buf)
{
printk(KERN_ERR PRCP "dyn1_name(%d, %d, %d)\n", index, (int)context, (int)name_buf);
	if ((index < 0) || (index >= DYNDIRNUM))
		return -ENOENT;
	if (((int)context) != 42)
		printk(KERN_ERR PRCP "dyn1_name expected 42 context\n");
	
	return sprintf(name_buf, "%d", index);
}
void *dyn1_context(int index,
		   void *old_context, 
		   void *data)
{
	int c = (int) old_context;
printk(KERN_ERR PRCP "dyn1_context(%d, %d)\n", index, (int)old_context);
	if ((index < 0) || (index >= DYNDIRNUM))
		return NULL;
	if (c != 42)
		printk(KERN_ERR PRCP "dyn1_context expected 42 context\n");

	dyndirc = c-index;
	return (void*) dyndirc;
}
void dyn1_leave(void *old_context,
		void *new_context, 
		void *data)
{
printk(KERN_ERR PRCP "dyn1_leave(%d, %d)\n", (int)old_context, (int)new_context);
	if (((int)old_context) != 42)
		printk(KERN_ERR PRCP "dyn1_leave expected 42 old_context\n");
	if (((int)new_context) != dyndirc)
		printk(KERN_ERR PRCP "dyn1_leave expected %d new_context\n",
		       dyndirc);
}
static void *add1_enter(void *old_context, 
			void *data)
{
	return (void*) (((int)old_context) + 1);
}
static void add1_leave(void *old_context,
		       void *new_context, 
		       void *data) 
{
	if (((int)old_context) != (((int)new_context)-1))
		printk(KERN_ERR PRCP "add1_leave check failed\n");
}

int dyn2_1_name(int index,
	        void *context, 
		void *data,
	        char *name_buf)
{
printk(KERN_ERR PRCP "dyn2_1_name(%d, %d, %d)\n", index, (int)context, (int)name_buf);
	if ((index < 0) || (index >= DYNDIRNUM))
		return -ENOENT;
	if (((int)context) != 42)
		printk(KERN_ERR PRCP "dyn2_1_name expected 42 context\n");
	
	return sprintf(name_buf, "p1_%d", index);
}
int dyn2_2_name(int index,
	        void *context, 
		void *data,
	        char *name_buf)
{
printk(KERN_ERR PRCP "dyn2_2_name(%d, %d, %d)\n", index, (int)context, (int)name_buf);
	if ((index < 0) || (index >= DYNDIRNUM))
		return -ENOENT;
	if (((int)context) != 42)
		printk(KERN_ERR PRCP "dyn2_2_name expected 42 context\n");
	
	return sprintf(name_buf, "p2_%d", index);
}
void *dyn2_2_context(int index,
		     void *old_context, 
		     void *data)
{
	int c = (int) old_context;
printk(KERN_ERR PRCP "dyn2_2_context(%d, %d)\n", index, (int)old_context);
	if ((index < 0) || (index >= DYNDIRNUM))
		return NULL;
	if (c != 42)
		printk(KERN_ERR PRCP "dyn2_2_context expected 42 context\n");

	return (void*) index;
}
int odd_num(void *context, 
	    void *data) 
{
printk(KERN_ERR PRCP "odd_num(%d)\n", (int)context);
	if ((((int)context) % 2) == 0)
		return 0;
	else 
		return 1;
}
int odd_name(int index,
	      void *context, 
	     void *data,
	      char *name_buf)
{
printk(KERN_ERR PRCP "odd_name(%d, %d, %d)\n", index, (int)context, (int)name_buf);
	if ((index < 0) || (index >= DYNDIRNUM))
		return -ENOENT;
	if ((((int)context) % 2) == 0)
		return -1;
	else 
		return sprintf(name_buf, "odd");

}

const char *enum_values[] = {"zero", "one", "two", "three", "four", "five"};

struct proc_dir_entry *root, *sims, *sims2, *simi, *simi2, *sime, *sime2;
struct proc_dir_entry *dir1, *dir2, *dir3, *dir1a, *dir1b, *dir1aa, *dir1aaa;
struct proc_dir_entry *file1_1, *file1_2, *file1a, *file1b, *file1aa, 
	*file1aaa_1, *file1aaa_2, *file2as, *file2bi, *file2ce;
struct proc_dir_entry *dyndir1, *dynsub, *dynview, *dynsubsub, *dynsubview;
struct proc_dir_entry *dyndir2, *dynsub2_1, *dynsub2_2, *dynsub2_1view, 
	              *dynsub2_2view, *dynoddcontext, *dynoddview;


void test_simple_cases()
{
	sims = proc_ov_register_string(root, "simple_string", 
				       write_simple_string);
	sims2 = proc_ov_register_string(NULL, "simple_string2", 
					write_simple_string);
	simi = proc_ov_register_int(root, "simple_int", write_simple_int);
	simi2 = proc_ov_register_int(NULL, "simple_int2", write_simple_int);
	sime = proc_ov_register_enum(root, "simple_enum", enum_values, 
				     write_simple_enum);
	sime2 = proc_ov_register_enum(NULL, "simple_enum2", enum_values, 
				      write_simple_enum);
}

void test_context_dirs()
{
	dir1 = proc_ov_register_dir(root, "dir1", simple_context_enter,
				    simple_context_leave);
	dir2 = proc_ov_register_dir(root, "dir2", simple_context_enter,
				    simple_context_leave);
	dir3 = proc_ov_register_dir(root, "dir3", NULL, NULL);
	dir1a = proc_ov_register_dir(dir1, "sub1", NULL, NULL);
	dir1b = proc_ov_register_dir(dir1, "sub2", subenter42, subleave42);
	dir1aa = proc_ov_register_dir(dir1a, "subsub", subenter42, subleave42);
	dir1aaa = proc_ov_register_dir(dir1aa, "subsubsub", subenter43, subleave43);
}

void test_context_files()
{
	file1_1 = proc_ov_register_int(dir1, "contextint", get_context_int);
	file1_2 = proc_ov_register_int(dir1, "is17", write_17);
	file1a = proc_ov_register_int(dir1a, "contextint2", get_context_int);
	file1b = proc_ov_register_int(dir1b, "contextint3", get_context_int);
	file1aa = proc_ov_register_int(dir1aa, "contextint4", get_context_int);
	file1aaa_1 = proc_ov_register_int(dir1aaa, "contextint5", get_context_int);
	file1aaa_2 = proc_ov_register_int(dir1aaa, "is17", write_17);
	file2as = proc_ov_register_string(dir2, "sayHello", write_hello);
	file2bi = proc_ov_register_int(dir2, "is17", write_17);
	file2ce = proc_ov_register_enum(dir2, "sayTwo", enum_values, write_enum_two);
}



void test_dyn_dirs()
{
	dyndir1 = proc_ov_register_dir(NULL, "dyndir1", simple_context_enter,
				       simple_context_leave);
	dynsub = proc_ov_register_dynamic_dir(dyndir1, DYNDIRNUM, dyn1_num, 
					      dyn1_name, dyn1_context, 
					      dyn1_leave);
	dynview = proc_ov_register_int(dynsub, "contextint", get_context_int);
	dynsubsub = proc_ov_register_dir(dynsub, "sub", add1_enter, add1_leave);
	dynsubview = proc_ov_register_int(dynsubsub, "contextint2", 
					  get_context_int);

	dyndir2 = proc_ov_register_dir(NULL, "dyndir2", simple_context_enter,
				       simple_context_leave);
	dynsub2_1 = proc_ov_register_dynamic_dir(dyndir2, DYNDIRNUM, dyn1_num,
		            dyn2_1_name, dyn1_context, dyn1_leave);
	dynsub2_2 = proc_ov_register_dynamic_dir(dyndir2, DYNDIRNUM, dyn1_num,
		            dyn2_2_name, dyn2_2_context, NULL);
	dynsub2_1view = proc_ov_register_int(dynsub2_1, "contextints1", 
					     get_context_int);
	dynsub2_2view = proc_ov_register_int(dynsub2_2, "contextints2", 
					     get_context_int);
	dynoddcontext = proc_ov_register_dynamic_dir(dynsub2_1, 1, odd_num,
						odd_name, NULL, NULL);
	dynoddview = proc_ov_register_int(dynoddcontext, "view", 
					  get_context_int);
}

int init_module(void)
{
	struct proc_dir_entry *root2, *root3;
	printk(KERN_DEBUG "test_proc_ov: test module initializing\n");
	
	root  = proc_ov_get_entry("/proc/");
	root2 = proc_ov_get_entry("/proc");
	root3 = proc_ov_get_entry("/");
	if ((root != root2) || (root != root3)) 
		printk(KERN_ERR "test_proc_ov: different root results");

	test_simple_cases();
	test_context_dirs();
	test_context_files();
	test_dyn_dirs();

	printk(KERN_DEBUG "test_proc_ov: test module installed\n");
	return 0;
}

void cleanup_module(void)
{
	#define r(x) if(!x) printk(KERN_DEBUG PRCP " entry failed: " #x "\n"); else proc_ov_remove(x);
	r(sims);
	r(sims2);
	r(simi);
	r(simi2);
	r(sime);
	r(sime2);

	r(file1_1);
	r(file1_2);
	r(file1a);
	r(file1b);
	r(file1aa);
	r(file1aaa_1);
	r(file1aaa_2);
	r(file2as);
	r(file2bi);
	r(file2ce);

	r(dir1);
	r(dir2);
	r(dir3);
	r(dir1a);
	r(dir1b);
	r(dir1aa);
	r(dir1aaa);

	r(dyndir1);
	r(dynsub);
	r(dynview);
	r(dynsubsub);
	r(dynsubview);

	r(dyndir2);
	r(dynsub2_1);
	r(dynsub2_2);
	r(dynsub2_1view);
	r(dynsub2_2view);
	r(dynoddcontext);
	r(dynoddview);

	printk(KERN_DEBUG "proc_ov test module removed\n");
}
