/*
 * Test for the functions & macros of vector.h
 */

#include "../../linux/include/linux/vector.h"
#include <assert.h>

static int numa[] = { 10, 6, 3, 7, 2, 9, 5, 8, 4, 1, -1};
static int *nums = numa;

static struct vector v1;

static void compare() {
	int i = 0;
	int *ip;
	while (nums[i] >= 0) {
		assert(VECTOR_GET_ELEMENT_AT(&v1, i, int) == nums[i]);
		assert(*VECTOR_GET_ELEMENTPTR_AT(&v1, i, int) == nums[i]);
		assert(VECTOR_GET_ARRAY(&v1, int)[i] == nums[i]);
		i++;
	}
	assert(VECTOR_GET_SIZE(&v1) == i);
	
	i = 0;
	VECTOR_FOR_EACH(&v1, ip, int) {
		assert(*ip == nums[i]);
		i++;
	}
}

static void fixed_tests() {
	int i;
	INIT_VECTOR(&v1, int, 0, 0);

	assert(VECTOR_GET_SIZE(&v1) == 0);

	i = 0;
	while (nums[i] >= 0) {
		VECTOR_ADD(&v1, &nums[i], int, 0);
		i++;
	}
	assert(VECTOR_GET_SIZE(&v1) == i);


	compare();
	
	VECTOR_REMOVE_INDEX(&v1, 0, int, 0);
	nums++;
	compare();

	VECTOR_REMOVE_ELEMENT(&v1, &nums[0], int, 0);
	nums++;
	compare();
	
	VECTOR_REMOVE_INDEX(&v1, 7, int, 0);
	nums[7] = -1;
	compare();

	VECTOR_REMOVE_ELEMENT(&v1, &nums[6], int, 0);
	nums[6] = -1;
	compare();

	VECTOR_REMOVE_INDEX(&v1, 4, int, 0);
	nums[4] = nums[5];
	nums[5] = -1;
	compare();

	VECTOR_REMOVE_ELEMENT(&v1, &nums[3], int, 0);
	nums[3] = nums[4];
	nums[4] = -1;
	compare();

	VECTOR_REMOVE_INDEX_REORDER(&v1, 0, int, 0);
	nums[0] = nums[3];
	nums[3] = -1;
	compare();

	VECTOR_REMOVE_INDEX_REORDER(&v1, 1, int, 0);
	nums[1] = nums[2];
	nums[2] = -1;
	compare();

	VECTOR_REMOVE_ELEMENT_REORDER(&v1, &nums[0], int, 0);
	nums[0] = nums[1];
	nums[1] = -1;
	compare();

	VECTOR_REMOVE_ELEMENT_REORDER(&v1, &nums[0], int, 0);
	nums[0] = -1;
	compare();

	FREE_VECTOR(&v1);
}

static void resize_tests() 
{
	static int num = 5000;
	static int frac = 500;
	int i;
	INIT_VECTOR(&v1, int, 1, 0);

	for(i = 0; i < num; i++) {
		VECTOR_ADD(&v1, &i, int, 0);
	}
	assert(VECTOR_GET_SIZE(&v1) == i);
	for (i = 0; i < num; i++) {
		assert(VECTOR_GET_ELEMENT_AT(&v1, i, int) == i);
	}


	for (i = 0; i < frac; i++) {
		VECTOR_REMOVE_ELEMENT(&v1, &i, int, 0);
	}
	assert(VECTOR_GET_SIZE(&v1) == num-frac);
	for (i = frac; i < num; i++) {
		assert(VECTOR_GET_ELEMENT_AT(&v1, i-frac, int) == i);
	}

	for (i = 0; i < frac; i++) {
		VECTOR_REMOVE_INDEX(&v1, 0, int, 0);
	}
	assert(VECTOR_GET_SIZE(&v1) == num-(2*frac));
	for (i = 2*frac; i < num; i++) {
		assert(VECTOR_GET_ELEMENT_AT(&v1, i-(2*frac), int) == i);
	}

	for (i = num-frac; i < num; i++) {
		VECTOR_REMOVE_ELEMENT(&v1, &i, int, 0);
	}
	assert(VECTOR_GET_SIZE(&v1) == num-(3*frac));
	for (i = 2*frac; i < (num-frac); i++) {
		assert(VECTOR_GET_ELEMENT_AT(&v1, i-(2*frac), int) == i);
	}

	for (i = num-(3*frac)-1; i >= (num-4*frac); i--) {
		VECTOR_REMOVE_INDEX(&v1, i, int, 0);
	}
	assert(VECTOR_GET_SIZE(&v1) == num-(4*frac));
	for (i = 2*frac; i < (num-2*frac); i++) {
		assert(VECTOR_GET_ELEMENT_AT(&v1, i-(2*frac), int) == i);
	}

	for(i = 2*frac; i < num - (2*frac); i++)
		VECTOR_REMOVE_ELEMENT_REORDER(&v1, &i, int, 0);
	assert(VECTOR_GET_SIZE(&v1) == 0);

	FREE_VECTOR(&v1);
}

int main(int argc, char **argp) 
{
	printf("Testing vector.h\n");

	fixed_tests();
	resize_tests();

	printf("finished.. passed.\n");
	return 0;
}
